
# =====================================================================================================================
# **NOTICE**
# THIS IS A GENERATED FILE USING THE SOURCE FILE:
#  crab\common\lib\jvm\seeq-names\src\main\java\com\seeq\common\seeqnames\SeeqNames.java
# =====================================================================================================================

class SeeqNames:
    class FormulaType:
        scalar = 'Scalar'
        signal = 'Signal'
        sample = 'Sample'
        condition = 'Condition'
        capsule = 'Capsule'
        formula_item = 'FormulaItem'
        onnx_prediction_model = 'OnnxPredictionModel'
        onnx_anomaly_model = 'OnnxAnomalyModel'
    
    class AuditTrail:
        configuration_option = 'ConfigurationOption'
    
    class ImpactReport:
        impact_table_name = 'VCC_ImpactTable_v1'
        impact = 'impact'
        impact_name = 'Name'
        impact_description = 'Description'
        impact_owner = 'Impact Owner'
        impact_status = 'Status'
        impact_formula = 'Formula'
        impact_savings_type = 'Savings Type'
        impact_frequency = 'Frequency'
        impact_unit = 'Unit'
        impact_asset = 'Asset'
        impact_start_date = 'Impact Start Date'
        impact_end_date = 'Impact End Date'
        impact_quantify_amount = 'Impact Quantify Amount'
        assigned_to_worksheet = 'Worksheet Id'
        assigned_to_workbook = 'Workbook Id'
        impact_site = 'Impact Site'
        manager = 'Manager'
        impact_lessons_learned = 'Lesson Learned'
        impact_formula_param = 'Formula Param'
        impact_based_on = 'Based on'
    
    class GraphQL:
        class DataLoaderRegistry:
            creator_metadata_loader = 'CREATOR_METADATA_LOADER'
            item_metadata_loader = 'ITEM_METADATA_LOADER'
        
    
    class SeeqContextNameSpace:
        seeq_label_namespace = '__'
        usage_restricted_label_namespace = '__Seeq.'
        vantage_label_category_namespace = '__Vantage'
    
    class MlService:
        class Clustering:
            class Connection:
                name = 'Seeq ML Clustering'
                url = 'cluster'
                udf_package = 'seeqClustering'
                udf_name = 'cluster'
            
        
        class IsolationForest:
            class Connection:
                name = 'Seeq ML Isolation Forest'
                url = 'if'
                udf_package = 'seeqIF'
                udf_name = 'isolationForest'
            
        
        class SelfOrganizingMap:
            class Connection:
                name = 'Seeq ML Self-Organizing Map'
                url = 'som'
                udf_package = 'seeqSOM'
                udf_name = 'selfOrganizingMap'
            
        
        class PlsRegression:
            class Connection:
                name = 'Seeq ML PLS Regression'
                url = 'pls'
                udf_package = 'seeqPLS'
                udf_name = 'plsRegression'
            
        
    
    class Consumption:
        screenshot = 'Screenshot'
        topic = 'Topic'
    
    class Capabilities:
        class ScalingTables:
            data_id = 'ScalingTablesCapability'
            collective_name = 'scaling tables users'
        
        class LogViewer:
            data_id = 'LogViewerCapability'
            collective_name = 'log viewers'
        
        class DatasourceAdministration:
            data_id = 'DatasourceAdministrationCapability'
            collective_name = 'datasource administrators'
        
        class UserAdministration:
            data_id = 'UserAdministrationCapability'
            collective_name = 'user administrators'
        
        class AuditTrail:
            data_id = 'AuditTrailCapability'
            collective_name = 'audit trail viewers'
        
        class AnalyticsAdministration:
            data_id = 'AnalyticsAdministrationCapability'
            collective_name = 'analytics administrators'
        
        class Administration:
            data_id = 'Administrators'
            collective_name = 'administrators'
        
    
    class RequestIdPrefixes:
        unique = 'R^'
        screenshot = 'ScreenshotCapture^'
        invocation = 'Invocation^'
    
    class PdfSettings:
        class MarginUnits:
            inches = 'in'
            millimeters = 'mm'
            centimeters = 'cm'
            pixels = 'px'
        
        class PageSize:
            letter = 'Letter'
            legal = 'Legal'
            a5 = 'A5'
            a4 = 'A4'
            a3 = 'A3'
            tabloid = 'Tabloid'
        
        class PageOrientation:
            portrait = 'Portrait'
            landscape = 'Landscape'
        
        class Defaults:
            page_orientation = 'Portrait'
            page_size = 'Letter'
            page_margin = '0.5in'
        
    
    class Database:
        class Utils:
            empty_guid = '00000000-0000-0000-0000-000000000000'
        
        class Schemas:
            materialized_tables = 'materialized_tables'
        
    
    class MaterializedTables:
        datum_id_column = 'datum id'
        item_id_column = 'item id'
        event_data_column = 'event data'
        context_table_prefix = 'context_'
        created_at_column = 'created at'
        unique_row_id = 'Existing Row Identifier'
        csv_item_id_column = 'csv item id'
        class Parameters:
            property_name = 'propertyName'
            property_names = 'propertyNames'
            property_value = 'propertyValue'
            property_match_operator = 'propertyMatchOperator'
            property_match_operators = 'propertyMatchOperators'
            level = 'level'
            separator = 'separator'
            formula = 'formula'
            exclude_globally_scoped = 'excludeGloballyScoped'
            parameters_name = 'parameters'
            name = 'name'
            description = 'description'
            column_index_data_id = 'columnIndexDataId'
            scoped_to = 'scopedTo'
            identity_id = 'identityId'
            permission_string = 'permissionString'
            security_string = 'securityString'
            column_index = 'columnIndex'
            column_indexes = 'columnIndexes'
            column_index_item = 'columnIndexItem'
            column_index_value = 'columnIndexValue'
            column_index_source_text = 'columnIndexSourceText'
            column_index_replacement = 'columnIndexReplacement'
            column_index_unit_of_measure = 'columnIndexUnitOfMeasure'
            constant = 'constant'
            is_root = 'isRoot'
            item_type = 'itemType'
            item_types = 'itemTypes'
            regex = 'regex'
            replacement = 'replacement'
            remove_non_matches = 'removeNonMatches'
            lookup_value = 'lookupValue'
            variable_parameters = 'variableParameters'
            variable_parameter_strings = 'variableParameterStrings'
            prefix = 'prefix'
            suffix = 'suffix'
        
        class Rules:
            ancestor = 'ancestor'
            asset_creator = 'assetCreator'
            concat_columns = 'concatColumns'
            descendant = 'descendant'
            event_property = 'eventProperty'
            formula_creator = 'formulaCreator'
            get_item_property = 'getItemProperty'
            item_search = 'itemSearch'
            path = 'path'
            path_search = 'pathSearch'
            numeric_scalar_creator = 'numericScalarCreator'
            set_item_property = 'setItemProperty'
            string_scalar_creator = 'stringScalarCreator'
            text_extractor = 'textExtractor'
            text_replacement = 'textReplacement'
            tree_path_creator = 'treePathCreator'
            manager = 'manager'
            item_type = 'itemType'
            datasource = 'datasource'
            context_condition = 'contextCondition'
            class ScalarCreator:
                base_scalar_creator = 'scalarCreator'
                numeric_scalar_creator = 'numericScalarCreator'
                string_scalar_creator = 'stringScalarCreator'
            
            class Overrides:
                override = 'Override'
                boolean_override = 'booleanOverride'
                numeric_override = 'numericOverride'
                string_override = 'stringOverride'
                timestamp_override = 'timestampOverride'
                u_u_i_d_override = 'uuidOverride'
            
            class Constant:
                base_constant = 'constant'
                boolean_constant = 'booleanConstant'
                numeric_constant = 'numericConstant'
                string_constant = 'stringConstant'
                timestamp_constant = 'timestampConstant'
                u_u_i_d_constant = 'uuidConstant'
                boolean_lookup_constant = 'booleanLookupConstant'
                numeric_lookup_constant = 'numericLookupConstant'
                string_lookup_constant = 'stringLookupConstant'
                timestamp_lookup_constant = 'timestampLookupConstant'
                u_u_i_d_lookup_constant = 'uuidLookupConstant'
            
        
        class FormulaAcross:
            asset_column = 'Asset'
            path_column = 'Path'
            datasource_column = 'Datasource'
            value_uom_column = 'Value UOM'
        
    
    class Injection:
        system_state = 'SystemState'
        system_user = 'SystemUser'
        tree_definition = 'TreeDefinition'
    
    class Ignition:
        agent_name = 'Ignition Module Agent'
    
    class Connectors:
        seeq_developer_name = 'Seeq'
        unknown_third_party_developer_name = 'third-party developer'
        connections_json_key = 'Connections'
        datasource_managed_json_key = 'DatasourceManaged'
        class Connections:
            name_json_key = 'Name'
            datasource_id_json_key = 'Id'
            enabled_json_key = 'Enabled'
            indexing_json_key = 'Indexing'
            frequency_json_key = 'Frequency'
            next_indexing_at_json_key = 'Next'
            on_startup_and_config_change_json_key = 'OnStartupAndConfigChange'
            transforms_json_key = 'Transforms'
            max_concurrent_requests_json_key = 'MaxConcurrentRequests'
            max_results_per_request_json_key = 'MaxResultsPerRequest'
            field_configurations_json_key = 'FieldConfigurations'
            class SyncStatus:
                unknown = 'SYNC_UNKNOWN'
                initializing = 'SYNC_INITIALIZING'
                in_progress = 'SYNC_IN_PROGRESS'
                archiving_deleted_items = 'SYNC_ARCHIVING_DELETED_ITEMS'
                complete = 'SYNC_COMPLETE'
                success = 'SYNC_SUCCESS'
                failed = 'SYNC_FAILED'
            
            class Status:
                connecting = 'CONNECTING'
                connected = 'CONNECTED'
                disconnected = 'DISCONNECTED'
                disabled = 'DISABLED'
            
            class SyncMode:
                none = 'NONE'
                incremental = 'INCREMENTAL'
                full = 'FULL'
            
            class SyncResult:
                in_progress = 'IN_PROGRESS'
                success = 'SUCCESS'
                failed = 'FAILED'
                interrupted = 'INTERRUPTED'
            
            class MultivaluedSamplesBehavior:
                keep_last = 'Keep Last'
                stagger = 'Stagger'
            
        
    
    class Agents:
        agent_keys_folder_name = 'agent-keys'
        connectors_json_key = 'Connectors'
        class WebSockets:
            ping_prefix = 'SEEQ-WS-PING-MSG-'
        
        class Status:
            connecting = 'CONNECTING'
            connected = 'CONNECTED'
            disconnected = 'DISCONNECTED'
            pre_provisioned = 'PREPROVISIONED'
        
        class Connectors:
            name_json_key = 'Name'
            enabled_json_key = 'Enabled'
        
    
    class TreeType:
        seeq_workbench = 'Seeq Workbench'
    
    class OData:
        export_entity_set_suffix = '_DataSet'
        o_data_v4_service_suffix = '.svc/'
    
    class CleanUpJobsGroup:
        annotation_images = 'AnnotationOrphanedImageCleanerJob'
        worksheet_maintenance = 'WorksheetMaintenanceJob'
    
    class JobGroup:
        screenshot = 'ScreenshotCapture'
        date_range = 'DateRangeJobs'
        report = 'ReportJobs'
        notebook = 'NotebookJobs'
        clean_up = 'CleanUpJobs'
        report_template = 'ReportTemplateJobs'
        condition_monitor = 'ConditionMonitorNotifications'
        worksheet_thumbnail = 'WorksheetThumbnail'
        item_finder = 'ItemFinder'
    
    class Logging:
        migrations_complete_file = 'migrations_complete_'
        class RemoteAgents:
            m_d_c_filename_key = 'agentlogfilename'
        
        class Appserver:
            m_d_c_filename_key = 'logfilename'
            default_file = 'appserver'
            startup_file = 'appserver-startup'
            upgrade_file = 'appserver-upgrade'
            async_upgrade_file = 'appserver-async-upgrade'
            licensing_file = 'appserver-licensing'
            requests = 'appserver-requests'
            request_monitors = 'appserver-requestmonitors'
            units = 'appserver-units'
            users = 'appserver-users'
            class IntervalAnalysis:
                fetch_critic = 'appserver-intervalanalysis-fetchcritic'
            
        
    
    class RemoteDatasourceClasses:
        l_d_a_p = 'LDAP'
        o_auth2 = 'OAuth 2.0'
        windows_auth = 'Windows Auth'
        a_d_x = 'ADX'
        a_p_r_m = 'AspenTech APRM'
        big_query = 'BigQuery'
        chaos_monkey = 'Chaos Monkey'
        cognite = 'Cognite'
        c_o_n_n_e_c_t_data_services = 'CONNECT Data Services'
        cyg_net = 'CygNet'
        example_data = 'Time Series CSV Files'
        add_on_calc_python = 'Add-on Calculation Python'
        add_on_calc_webhook = 'Add-on Calculation Webhook'
        add_on_calc_excel = 'Add-on Calculation Excel VBA'
        file_folders = 'FileFolders'
        g_e_proficy = 'GE Proficy'
        g_p_m = 'GreenPowerMonitor'
        i_b_a_h_d = 'IbaHD'
        ignition = 'Ignition Gateway'
        influx_d_b = 'InfluxDB'
        influx_d_b2 = 'InfluxDB 2'
        i_p21 = 'AspenTech IP21'
        met_net = 'MetNet'
        n_o_a_a_weather = 'NOAA Weather Service'
        omnia = 'Equinor Omnia Plant'
        o_p_c_h_d_a = 'OPC-HDA'
        o_p_c_u_a = 'OPC-UA'
        o_s_isoft_a_f = 'OSIsoft AF'
        o_s_isoft_a_f_event_frames = 'OSIsoft AF Event Frames'
        o_s_isoft_p_i = 'OSIsoft PI'
        seeq_to_seeq = 'Seeq to Seeq'
        s_q_l = 'SQL'
        stress_test = 'Stress Test'
        time_stream = 'TimeStream'
        tree_file = 'Tree File'
        w_i_t_s_m_l = 'WITSML'
        wonderware = 'Wonderware'
        x_h_q = 'XHQ'
    
    class SeeqSaasAdminDatasource:
        datasource_class = 'OAuth 2.0'
        datasource_id = 'Seeq Saas Admin'
        datasource_name = 'Seeq Saas Admin'
    
    class IdentityProviderDatasourceClasses:
        l_d_a_p = 'LDAP'
        o_auth2 = 'OAuth 2.0'
        seeq_auth = 'Auth'
        windows_auth = 'Windows Auth'
    
    class LocalDatasources:
        class MaterializedTables:
            datasource_class = 'Seeq Materialized Tables'
        
        class SeeqDataLab:
            datasource_class = 'Seeq Data Lab'
        
        class SeeqMetadataItemStorage:
            datasource_class = 'Seeq Metadata'
            datasource_id = 'Seeq Metadata'
            datasource_name = 'Seeq Metadata'
        
        class SeeqCalculationItemStorage:
            datasource_class = 'Seeq Calculations'
            datasource_id = 'Seeq Calculations'
            datasource_name = 'Seeq Calculations'
        
        class Monitors:
            datasource_class = 'Monitors'
            datasource_id = 'monitors'
            datasource_name = 'Seeq Monitors'
        
        class Authentication:
            datasource_class = 'Auth'
            datasource_id = 'Seeq'
            datasource_name = 'Seeq'
            password_special_characters = '!@#$%^&*()_+-=[]{}|;:,.<>?/'
        
        class FileSignalStorage:
            datasource_class = 'Seeq - Signal Datasource'
            datasource_id = 'Signal Datasource'
            datasource_name = 'Seeq Signal Datasource'
        
        class ContextDatumStorage:
            datasource_class = 'Context Datum Storage'
            datasource_id = 'Context Datum Storage'
            datasource_name = 'Context Datum Storage'
        
        class PostgresDatums:
            datasource_class = 'Postgres Datums'
            datasource_id = 'default'
            datasource_name = 'Seeq Conditions Database'
        
        class Datafile:
            datasource_class = 'Imported CSV Files'
            datasource_id = 'Imported CSV Files'
            datasource_name = 'Imported CSV Files'
        
    
    class SwaggerApiNames:
        agents = 'Agents'
        add_ons = 'AddOns'
        audit = 'Audit'
        auth = 'Auth'
        assets = 'Assets'
        folders = 'Folders'
        networks = 'Networks'
        items = 'Items'
        reports = 'Reports'
        jobs = 'Jobs'
        logs = 'Logs'
        trees = 'Trees'
        sample_series = 'SampleSeries'
        signals = 'Signals'
        capsules = 'Capsules'
        conditions = 'Conditions'
        metrics = 'Metrics'
        formulas = 'Formulas'
        functions = 'Functions'
        requests = 'Requests'
        tables = 'Tables'
        workbooks = 'Workbooks'
        worksheets = 'Worksheets'
        export = 'Export'
        annotations = 'Annotations'
        datasources = 'Datasources'
        datafiles = 'Datafiles'
        users = 'Users'
        user_groups = 'UserGroups'
        access_keys = 'AccessKeys'
        usage = 'Usage'
        scalars = 'Scalars'
        subscriptions = 'Subscriptions'
        system = 'System'
        import_ = 'Import'
        monitors = 'Monitors'
        projects = 'Projects'
        content = 'Content'
        report_templates = 'ReportTemplates'
        condition_monitors = 'ConditionMonitors'
        plugins = 'Plugins'
        displays = 'Displays'
        display_templates = 'DisplayTemplates'
        notification_configurations = 'NotificationConfigurations'
        label = 'Label'
        context = 'Context'
        table_definitions = 'TableDefinitions'
        models = 'Models'
        table_overrides = 'TableOverrides'
        graph_q_l = 'GraphQL'
        unstable = 'Unstable'
        scim = 'SCIM'
    
    class Channels:
        agents_status = 'agents-status'
        datasources_status = 'datasources-status'
        broadcast = 'broadcast'
        live_screenshot = 'live-screenshot'
        requests_progress = 'requests-progress'
        workbook_channel = 'workbook'
        locking_channel = 'workbook-locking'
        thumbnail_channel = 'workbook-thumbnail'
        workstep_channel = 'workstep'
        report_update_channel = 'report-update'
        table_materializer_channel = 'table-materializer'
    
    class API:
        ace = '/ace'
        add_ons = '/add-ons'
        add_on_tools = '/addOnTools'
        agents = '/agents'
        audit = '/audit'
        auth = '/auth'
        assets = '/assets'
        gen_a_i = '/genai'
        folders = '/folders'
        networks = '/networks'
        items = '/items'
        item_finders = '/finders'
        reports = '/reports'
        jobs = '/jobs'
        logs = '/logs'
        batch = '/batch'
        trees = '/trees'
        asset_trees = '/trees/assets'
        signals = '/signals'
        samples = '/samples'
        plot_samples = '/plot-samples'
        capsules = '/capsules'
        conditions = '/conditions'
        condition_monitors = '/condition-monitors'
        metrics = '/metrics'
        thresholds = '/thresholds'
        formulas = '/formulas'
        docs = '/docs'
        functions = '/functions'
        packages = '/packages'
        scalars = '/scalars'
        properties = '/properties'
        requests = '/requests'
        tables = '/tables'
        workbooks = '/workbooks'
        worksheets = '/worksheets'
        worksteps = '/worksteps'
        export = '/export'
        annotations = '/annotations'
        datasources = '/datasources'
        datafiles = '/datafiles'
        users = '/users'
        user_groups = '/usergroups'
        access_keys = '/accesskeys'
        o_data = '/odata.svc'
        o_data_v4 = '/odata'
        system = '/system'
        import_ = '/import'
        status = '/status'
        shutdown = '/shutdown'
        screenshots = '/screenshots'
        subscriptions = '/subscriptions'
        monitors = '/system/monitors'
        configuration = '/system/configuration'
        installer = '/system/installer'
        projects = '/projects'
        content = '/content'
        report_templates = '/report-templates'
        date_ranges = '/date-ranges'
        asset_selections = '/asset-selections'
        plugins = '/plugins'
        displays = '/displays'
        display_templates = '/display-templates'
        usage = '/usage'
        notification_configurations = '/notification-configurations'
        labels = '/labels'
        context = '/context'
        context_comments = '/comments'
        context_labels = '/labels'
        context_opaque = '/opaque'
        context_numeric = '/numeric'
        table_definitions = '/table-definitions'
        models = '/models'
        graph_q_l = '/graphql'
        cache = '/cache'
        scim_users = 'Users'
        scim_groups = 'Groups'
        scim_token = 'token'
        class ErrorMessages:
            attempted_to_set_scope_on_a_globally_scoped_item = 'Attempted to set scope on a globally scoped item'
            cancelled_due_to_too_many_retries = 'The request was canceled due to being retried too many times'
        
        class AddOnToolLinkType:
            window = 'window'
            tab = 'tab'
            none = 'none'
        
        class AddOnToolLaunchLocation:
            homescreen = 'homescreen'
            tools_pane = 'toolsPane'
        
        class PropertyIds:
            request_information = 'com.seeq.requestinformation'
        
        class Cookies:
            auth = 'sq-auth'
            csrf = 'sq-csrf'
        
        class QueryParams:
            admin = 'admin'
            client_interactive_session_id = 'interactiveId'
            creator = 'creator'
            name = 'name'
            request_id = 'requestId'
            content_count = 'contentCount'
        
        class Headers:
            accept_ranges = 'Accept-Ranges'
            async_ = 'x-sq-async'
            async_job = 'x-sq-async-job'
            auth = 'x-sq-auth'
            identity_path = 'x-sq-identity-path'
            content_range = 'Content-Range'
            csrf = 'x-sq-csrf'
            forwarded_for = 'x-forwarded-for'
            interactive_session_id = 'Session-Id'
            pool_id = 'x-sq-pool-id'
            pool_connected_on_degraded_state = 'x-sq-pool-connected-on-degraded-state'
            range = 'Range'
            request_id = 'x-sq-request-id'
            request_origin = 'x-sq-origin'
            request_origin_u_r_l = 'x-sq-origin-url'
            request_origin_label = 'x-sq-origin-label'
            server = 'Server'
            server_meters = 'Server-Meters'
            server_timing = 'Server-Timing'
            expect = 'Expect'
            date_range_start = 'x-sq-daterange-start'
            date_range_end = 'x-sq-daterange-end'
            class Timings:
                datasource = 'Datasource'
                cache = 'Cache'
                processing = 'Processing'
                g_c = 'GC'
                request_queue = 'Request Queue'
                calc_engine_queue = 'Calc Engine Queue'
                seeq_database = 'Seeq Database'
                total = 'Total'
            
            class Meters:
                datasource_samples_read = 'Datasource Samples Read'
                datasource_capsules_read = 'Datasource Capsules Read'
                cache_samples_read = 'Cache Samples Read'
                cache_capsules_read = 'Cache Capsules Read'
                cache_in_memory_samples_read = 'Cache In-Memory Samples Read'
                cache_in_memory_capsules_read = 'Cache In-Memory Capsules Read'
                database_items_read = 'Database Items Read'
                database_relationships_read = 'Database Relationships Read'
            
        
        class Flags:
            all_properties = '@allProperties'
            exclude_globally_scoped = '@excludeGloballyScoped'
            include_unsearchable = '@includeUnsearchable'
        
        class TranslationKeyPrefixes:
            capabilities = 'API.CAPABILITIES.'
        
    
    class UnitTest:
        timestamp1 = '1990-03-07T09:58:20.888888Z'
        timestamp2 = '1990-04-08T10:59:21.999999Z'
        non_existent_guid = '1CB64296-FC3E-4221-9A7F-947A0AB99807'
    
    class Edges:
        owns = 'owns'
        created = 'created'
        installed = 'installed'
        contains = 'contains'
        folder_contains = 'folder contains'
        member_of = 'member of'
        interested_in = 'interested in'
        reports_on = 'reports on'
        replies_to = 'replies to'
        recently_accessed = 'recently accessed'
        has_worksheet = 'has worksheet'
        has_workstep = 'has workstep'
        current_workstep = 'current workstep'
        template_workstep = 'template workstep'
        has_parameter = 'has parameter'
        uses_function = 'uses function'
        associated_to = 'associated to'
        # DEPRECATED, Deprecated. See SeeqNames.java for more info
        access_for = 'access for'
        # DEPRECATED, Deprecated. See SeeqNames.java for more info
        has_ace = 'has access control entry'
        has_pinned = 'has pinned'
        measuring = 'measuring'
        has_aggregation_signal = 'has aggregation signal'
        has_aggregation_condition = 'has aggregation condition'
        has_aggregation_formula_function = 'has aggregation formula function'
        bounded_by = 'bounded by'
        has_threshold = 'has threshold'
        has_condition_threshold = 'has condition threshold'
        home_folder = 'home folder'
        has_manager = 'has manager'
        package_contains_u_d_f = 'package contains user defined function'
        package_contains_doc = 'package contains formula doc'
        documented_by = 'documented by'
        agent_contains_connector = 'agent contains connector'
        connection_created_by_connector = 'connection created by connector'
        renders = 'renders'
        runs_notebook = 'runs notebook'
        source_worksheet = 'source worksheet'
        source_workstep = 'source workstep'
        has_asset = 'has asset'
        has_asset_selection = 'has asset selection'
        has_asset_selections = 'has asset selections'
        has_condition = 'has condition'
        has_item_finder = 'has item finder'
        has_content = 'has content'
        has_date_range = 'has date range'
        has_date_ranges = 'has date ranges'
        has_designated_scim_user = 'has designated scim user'
        has_scheduled_notebook = 'has scheduled notebook'
        has_template = 'has template'
        subscribed_to = 'subscribed to'
        contains_context_for = 'contains context for'
    
    class Types:
        asset = 'Asset'
        asset_selection = 'AssetSelection'
        access_control_entry = 'AccessControlEntry'
        stored_signal = 'StoredSignal'
        datasource = 'Datasource'
        datafile = 'Datafile'
        calculated_signal = 'CalculatedSignal'
        formula_item = 'FormulaItem'
        signal = 'Signal'
        tree_definition = 'TreeDefinition'
        identity = 'Identity'
        user = 'User'
        user_group = 'UserGroup'
        access_key = 'AccessKey'
        system_state = 'SystemState'
        export = 'Export'
        project_link = 'ProjectLink'
        add_on_tool = 'AddOnTool'
        base_annotation = 'BaseAnnotation'
        report = 'Report'
        report_template = 'ReportTemplate'
        journal = 'Journal'
        reply = 'Reply'
        item_finder = 'ItemFinder'
        condition = 'Condition'
        condition_monitor = 'ConditionMonitor'
        capsule = 'Capsule'
        stored_condition = 'StoredCondition'
        calculated_condition = 'CalculatedCondition'
        scalar = 'ScalarItem'
        literal_scalar = 'LiteralScalar'
        calculated_scalar = 'CalculatedScalar'
        formula_function = 'FormulaFunction'
        aggregating_formula_function = 'AggregatingFormulaFunction'
        chart = 'Chart'
        user_defined_formula_function = 'UserDefinedFormulaFunction'
        formula_package = 'FormulaPackage'
        formula_doc = 'FormulaDoc'
        formula_package_doc = 'FormulaPackageDoc'
        threshold_metric = 'ThresholdMetric'
        display = 'Display'
        display_template = 'DisplayTemplate'
        folder = 'Folder'
        project = 'Project'
        content = 'Content'
        image_content = 'ImageContent'
        react_json_content = 'ReactJsonContent'
        date_range = 'DateRange'
        scheduled_notebook = 'ScheduledNotebook'
        add_on = 'AddOn'
        plugin = 'Plugin'
        connection = 'Connection'
        connector = 'Connector'
        agent = 'Agent'
        analysis = 'Analysis'
        topic = 'Topic'
        vantage = 'Vantage'
        worksheet = 'Worksheet'
        workstep = 'Workstep'
        item = 'Item'
        item_with_owner = 'ItemWithOwner'
        type_definition = 'TypeDefinition'
        table_definition = 'TableDefinition'
        version_history = 'VersionHistory'
    
    class TopicDocumentAttributes:
        data_seeq_content = 'data-seeq-content'
        data_seeq_content_pending = 'data-seeq-content-pending'
        data_seeq_date_range_id = 'data-date-range-id'
        data_seeq_date_range_format = 'data-date-range-format'
        data_seeq_date_range_content = 'data-date-range-content'
        data_seeq_content_percent_width = 'data-seeq-content-percent-width'
        data_seeq_content_no_margin = 'data-seeq-content-no-margin'
        data_seeq_content_border = 'data-seeq-content-border'
        data_seeq_content_height = 'data-seeq-content-height'
        data_seeq_content_width = 'data-seeq-content-width'
        data_seeq_content_property_overrides = 'data-seeq-content-property-overrides'
        data_seeq_asset_selection_id = 'data-asset-selection-id'
        data_seeq_asset_selection_depth_level = 'data-asset-selection-depth-level'
        data_seeq_iframe_content_url = 'data-iframe-content-url'
        data_seeq_iframe_content_height = 'data-iframe-content-height'
        data_seeq_iframe_content_width = 'data-iframe-content-width'
        data_seeq_markdown_content = 'data-markdown-content'
    
    class CapsuleProperties:
        start = 'Start'
        end = 'End'
        duration = 'Duration'
        is_uncertain = 'IsUncertain'
        is_bounded = 'IsBounded'
        middle = 'Middle'
        count = 'Count'
        value = 'Value'
        unit = 'Unit'
        distance = 'Distance'
        similarity = 'Similarity'
        standard_deviation = 'Standard Deviation'
        sample_period = 'Sample Period'
        reference_capsule = 'Reference Capsule'
        condition_id = 'Condition ID'
        capsule_id = 'Capsule ID'
        capsule_id_safe = 'Capsule SortKey'
        original_uncertainty = 'Original Uncertainty'
        is_suppressed = 'Is Suppressed'
    
    class Properties:
        guid = 'GUID'
        name = 'Name'
        id = 'ID'
        description = 'Description'
        unsearchable = 'Unsearchable'
        archived = 'Archived'
        archival_id = 'Archival ID'
        archived_at = 'Archived At'
        archived_reason = 'Archived Reason'
        dashboard = 'Dashboard'
        sync_token = 'Sync Token'
        number_format = 'Number Format'
        source_number_format = 'Source Number Format'
        string_format = 'String Format'
        source_string_format = 'Source String Format'
        storage_location = 'Storage Location'
        executor_id = 'Executor ID'
        creator_id = 'Creator ID'
        creator_first_name = 'Creator First Name'
        creator_last_name = 'Creator Last Name'
        historical_data_version = 'Historical Data Version'
        manual_cdc = 'Manual CDC'
        scoped_to = 'Scoped To'
        asset = 'asset'
        full_path = 'fullpath'
        locked = 'Locked'
        locked_metadata = 'Locked Metadata'
        import_item_type = 'Import Item Type'
        filename = 'Filename'
        name_row = 'Name Row'
        name_prefix = 'Name Prefix'
        name_suffix = 'Name Suffix'
        condition_name = 'Condition Name'
        first_data_row = 'First Data Row'
        description_row = 'Description Row'
        interpolation_method_row = 'Interpolation Method Row'
        maximum_interpolation_row = 'Maximum Interpolation Row'
        value_uom_row = 'Value Unit of Measure Row'
        field_delimiter = 'Field Delimiter'
        day_first_default = 'Day First Default'
        key_column_name = 'Key Column Name'
        key_column_index = 'Key Column Index'
        end_column_name = 'End Column Name'
        end_column_index = 'End Column Index'
        key_format = 'Key Format'
        time_zone = 'Time Zone'
        value_column_names = 'Value Column Names'
        value_column_indices = 'Value Column Indices'
        append = 'Append'
        validation_mode = 'Validation Mode'
        lenient_daylight_savings = 'Lenient Daylight Savings'
        datasource_class = 'Datasource Class'
        datasource_name = 'Datasource Name'
        datasource_id = 'Datasource ID'
        datasource_host_id = 'Datasource Host ID'
        data_id = 'Data ID'
        previous_data_id = 'Previous Data ID'
        stored_in_seeq = 'Stored In Seeq'
        connection_request_timeout = 'Connection Request Timeout'
        max_requests = 'Max Requests'
        windows_auth_mode = 'Windows Auth Mode'
        ldap_auth_mode = 'LDAP Auth Mode'
        oidc_issuer = 'Oidc Issuer'
        oidc_jwt_key_set = 'Oidc JWT Key Set'
        oidc_audience = 'Oidc Audience'
        username_password_compatible = 'Username/Password Compatible'
        username_human_readable = 'Username Human Readable'
        expect_duplicates_during_indexing = 'Expect Duplicates During Indexing'
        referenced_datasource_id = 'Referenced Datasource ID'
        referenced_datasource_class = 'Referenced Datasource Class'
        additional_groups_to_sync = 'Additional Groups To Sync'
        signal_location = 'Signal Location'
        condition_location = 'Condition Location'
        current_index_at = 'Current Index At'
        previous_index_at = 'Previous Index At'
        sync_result = 'Sync Result'
        indexing_schedule_supported = 'Indexing Schedule Supported'
        sync_mode = 'Sync Mode'
        items_archived_count = 'Items Archived Count'
        indexing_duration = 'Indexing Duration'
        provides_everyone_group_identities = 'Provides Everyone Group Identities'
        datasource_labels = 'Datasource Labels'
        datasource_vendor = 'Datasource Vendor'
        indexing_progress_timestamp = 'Indexing Progress Timestamp'
        asset_progress = 'Asset Progress'
        asset_count = 'Asset Count'
        previous_asset_count = 'Previous Asset Count'
        signal_progress = 'Signal Progress'
        signal_count = 'Signal Count'
        previous_signal_count = 'Previous Signal Count'
        condition_progress = 'Condition Progress'
        condition_count = 'Condition Count'
        previous_condition_count = 'Previous Condition Count'
        scalar_progress = 'Scalar Progress'
        scalar_count = 'Scalar Count'
        previous_scalar_count = 'Previous Scalar Count'
        relationship_progress = 'Relationship Progress'
        relationship_count = 'Relationship Count'
        previous_relationship_count = 'Previous Relationship Count'
        user_group_progress = 'UserGroup Progress'
        user_group_count = 'UserGroups Count'
        previous_user_group_count = 'Previous UserGroups Count'
        stale_metadata = 'Stale Metadata'
        interpolation_method = 'Interpolation Method'
        source_maximum_interpolation = 'Source Maximum Interpolation'
        override_maximum_interpolation = 'Override Maximum Interpolation'
        maximum_interpolation = 'Maximum Interpolation'
        value_uom = 'Value Unit Of Measure'
        source_value_uom = 'Source Value Unit Of Measure'
        display_unit = 'Display Unit'
        units_migrated = 'Units Migrated'
        maximum_duration = 'Maximum Duration'
        cache_enabled = 'Cache Enabled'
        precaching_enabled = 'Pre-Caching Enabled'
        uncertainty_override = 'Uncertainty Override'
        metadata_properties = 'Metadata Properties'
        last_key_written = 'Last Key Written'
        request_interval_clamp = 'Request Interval Clamp'
        old_cache_id = 'Old Cache ID'
        debounce_duration = 'Debounce Duration'
        debounce_window = 'Debounce Window'
        tombstoned_properties = 'Tombstoned Properties'
        multivalued_samples_behavior = 'Multivalued Samples Behavior'
        capsule_id_property = 'Capsule ID Property'
        stored_series_cache_version = 'Stored Series Cache Version'
        replace_capsule_properties = 'Replace Capsule Properties'
        uom = 'Unit Of Measure'
        formula_version = 'Formula Version'
        formula = 'Formula'
        swap_source_id = 'Swap Source ID'
        swap_source_uom = 'Source Value Unit Of Measure'
        swap_key = 'Swap Key'
        cache_value = 'CacheValue'
        value = 'Value'
        formula_parameters = 'FormulaParameters'
        is_modified_outside_table = 'Is Modified Outside Table'
        expected_provided_type = 'ExpectedProvidedType'
        version_string = 'Version String'
        creator_name = 'Creator Name'
        creator_contact_info = 'Creator Contact Info'
        title = 'Title'
        examples = 'Examples'
        search_keywords = 'Search Keywords'
        experimental_package_name = 'Experimental'
        seeq_package_name = 'Seeq'
        index_doc_name = 'index'
        aggregation_function = 'AggregationFunction'
        duration = 'Duration'
        period = 'Period'
        measured_item_maximum_duration = 'MeasuredItemMaximumDuration'
        measured_item_id = 'MeasuredItemId'
        metric_configuration_migration = 'MetricReconfigurationMigration'
        bounding_condition_maximum_duration = 'BoundingConditionMaximumDuration'
        neutral_color = 'NeutralColor'
        is_generated = 'Is Generated'
        tree_type = 'Tree Type'
        manually_added = 'Manually Added'
        created_by = 'Created By'
        asset_path = 'assetPath'
        column_type = 'columnType'
        original_mapping = 'originalMapping'
        parameter_to_column = 'parameterToColumn'
        swap_source_asset_id = 'Swap Source Asset ID'
        export_data = 'Export Data'
        is_capsule_view = 'Is Capsule View'
        is_export_capsule_table = 'Is Export Capsule Table'
        is_chain_view = 'Is Chain View'
        is_autoupdate_time_range = 'Is Autoupdate Time Range'
        grid_size = 'Grid Size'
        user_email = 'User Email'
        username = 'Username'
        first_name = 'First Name'
        last_name = 'Last Name'
        password_hash = 'Password Hash'
        password_hash_method = 'Password Hash Method'
        password_set_date = 'Password Set Date'
        last_login_at = 'Last Login At'
        failed_login_attempt_count = 'Failed Login Attempt Count'
        session_duration = 'Session Duration'
        access_key_validity_duration = 'Access Key Validity Duration'
        enabled = 'Enabled'
        workbench = 'Workbench'
        user_role = 'User Role'
        site = 'Site'
        user_language = 'userLanguage'
        require_password_reset_on_next_login = 'Require Password Reset On Next Login'
        agent_provisioning_status = 'Agent Provisioning Status'
        agent_provisioning_source_address = 'Agent Provisioning Source Address'
        agent_provisioning_requested_at = 'Agent Provisioning Requested At'
        password_reset_token_hash = 'Password Reset Token Hash'
        password_reset_token_expiration_at = 'Password Reset Token Expiration At'
        last_password_reset_request_at = 'Last Password Reset Request At'
        remote_group_editable = 'Remote Group Editable'
        remove_permissions = 'Remove Permissions'
        data = 'Data'
        created_at = 'Created At'
        updated_at = 'Updated At'
        order = 'Order'
        removed_worksheet_workstep_relationships = 'Removed Worksheet Workstep Relationships'
        document = 'Document'
        plain_text_document = 'Plain Text Document'
        discoverable = 'Discoverable'
        published_at = 'Published At'
        last_viewed_at = 'Last Viewed At'
        source_security_string = 'Source Security String'
        security_string = 'Security String'
        permissions_from_datasource = 'Permissions From Datasource'
        permission_inheritance_disabled = 'Permission Inheritance Disabled'
        asset_group_member = 'Asset Group Member'
        first_admin_created = 'First Admin Created'
        u_i_config = 'UIConfig'
        dirty_cache = 'Dirty Cache'
        unmodifiable = 'Unmodifiable'
        height = 'Height'
        width = 'Width'
        scale = 'Scale'
        timezone = 'Timezone'
        selector = 'Selector'
        last_screenshot_duration = 'Last Screenshot Duration'
        screenshot_warning = 'Screenshot Warning'
        summary_type = 'Summary Type'
        summary_value = 'Summary Value'
        hide_uncertainty = 'Hide Uncertainty'
        current_filename = 'Current Filename'
        last_result_hash = 'Last Result Hash'
        cache_id = 'Cache ID'
        asset_path_depth = 'Asset Path Depth'
        last_run_state = 'Last Run State'
        query_range_look_ahead = 'Query Range Look Ahead'
        webhook_url = 'Webhook Url'
        first_run_look_back = 'First Run Look Back'
        capsule_event_types = 'Capsule Event Types'
        finder_configurations = 'Finder Configurations'
        include_hierarchy = 'Include Hierarchy'
        resource_size = 'Resource Size'
        project_type = 'Project Type'
        startup_environment = 'Startup Environment'
        binary_data = 'Binary Data'
        last_run_at = 'Last Run At'
        # DEPRECATED, Deprecated. See SeeqNames.java for more info
        previous_run_time = 'Previous Run Time'
        scheduled_at = 'Scheduled At'
        total_run_time = 'Total Run Time'
        average_run_time = 'Average Run Time'
        cron_schedule = 'Cron Schedule'
        background = 'Background'
        condition_formula_now = 'Condition Formula Now'
        last_run_warnings = 'Last Run Warnings'
        last_run_errors = 'Last Run Errors'
        notebook_path = 'Notebook Path'
        cron_schedules_to_indices = 'Cron Schedules To Indices'
        label = 'Label'
        notify_on_skipped_execution = 'Notify On Skipped Execution'
        notify_on_automatic_unschedule = 'Notify On Automatic Unschedule'
        is_fixed_with = 'Is Fixed Width'
        margin_top = 'Margin Top'
        margin_bottom = 'Margin Bottom'
        margin_left = 'Margin Left'
        margin_right = 'Margin Right'
        orientation = 'Page Orientation'
        page_size = 'Page Size'
        is_ck_enabled = 'Is CK Enabled'
        froala_backup = 'Froala Backup'
        original_has_workstep_guid = 'Original HasWorkstep GUID'
        in_development = 'In Development'
        icon_class = 'Icon Class'
        target_url = 'Target URL'
        launch_location = 'Launch Location'
        link_type = 'Link Type'
        window_details = 'Window Details'
        reuse_window = 'Reuse Window'
        sort_key = 'Sort Key'
        add_on_identifier = 'Add-on Identifier'
        add_on_components = 'Add-on Components'
        html_template = 'HTML Template'
        plain_text_template = 'Plain Text Template'
        category = 'Category'
        identifier = 'Identifier'
        location = 'Location'
        entry_point = 'Entry Point'
        icon = 'Icon'
        last_update_time = 'Last Update Time'
        host = 'Host'
        options = 'Options'
        version = 'Version'
        json = 'Json'
        propagation_pending = 'Propagation Pending'
        datasource_managed = 'Datasource Managed'
        connection_id = 'ConnectionId'
        indexing_connection = 'Indexing Connection'
        pull_connection = 'Pull Connection'
        max_concurrent_requests = 'Max Concurrent Requests'
        max_results_per_requests = 'Max Results Per Requests'
        transforms = 'Transforms'
        type = 'Type'
        indexing_frequency = 'Indexing Frequency'
        next_scheduled_index_at = 'Next Scheduled Index At'
        disabled_at = 'Disabled At'
        disabled_by = 'Disabled By'
        indexing_on_startup_and_config_change = 'Indexing On Startup And Config Change'
        column_definitions = 'Column Definitions'
        batch_action = 'Batch Action'
        table_type = 'Table Type'
        is_context_condition = 'Is Context Condition'
        versioned_item_type = 'Versioned Item Type'
        versions = 'Versions'
        inputs = 'Inputs'
        result_type = 'Result Type'
    



